# Orca
#
# Copyright 2011-2025 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for chat."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .chat_presenter import ChatPresenter


def get_commands(owner: ChatPresenter) -> list[Command]:
    """Returns commands for chat."""

    return [
        KeyboardCommand(
            "chat_toggle_room_name_prefix",
            owner.toggle_prefix,
            owner.GROUP_LABEL,
            cmdnames.CHAT_TOGGLE_ROOM_NAME_PREFIX,
        ),
        KeyboardCommand(
            "chat_toggle_buddy_typing",
            owner.toggle_buddy_typing,
            owner.GROUP_LABEL,
            cmdnames.CHAT_TOGGLE_BUDDY_TYPING,
        ),
        KeyboardCommand(
            "chat_toggle_message_histories",
            owner.toggle_message_histories,
            owner.GROUP_LABEL,
            cmdnames.CHAT_TOGGLE_MESSAGE_HISTORIES,
        ),
        KeyboardCommand(
            "chat_previous_message",
            owner.present_previous_message,
            owner.GROUP_LABEL,
            cmdnames.CHAT_PREVIOUS_MESSAGE,
        ),
        KeyboardCommand(
            "chat_next_message",
            owner.present_next_message,
            owner.GROUP_LABEL,
            cmdnames.CHAT_NEXT_MESSAGE,
        ),
    ]



# Mouse presenter for Orca
#
# Copyright 2008 Eitan Isaacson
# Copyright 2016 Igalia, S.L.
# Copyright 2026 SUSE LLC.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for mouse presentation and interaction."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .mouse_presenter import MousePresenter


def get_commands(owner: MousePresenter) -> list[Command]:
    """Returns commands for mouse presentation and interaction."""

    kb_orca_kp_divide = keybindings.KeyBinding(
        "KP_Divide",
        keybindings.ORCA_MODIFIER_MASK,
    )
    kb_unmodified_kp_divide = keybindings.KeyBinding(
        "KP_Divide",
        keybindings.NO_MODIFIER_MASK,
    )
    kb_unmodified_kp_multiply = keybindings.KeyBinding(
        "KP_Multiply",
        keybindings.NO_MODIFIER_MASK,
    )
    kb_orca_9 = keybindings.KeyBinding("9", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_7 = keybindings.KeyBinding("7", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_8 = keybindings.KeyBinding("8", keybindings.ORCA_MODIFIER_MASK)

    return [
        KeyboardCommand(
            "toggleMouseReviewHandler",
            owner.toggle_mouse_review,
            owner.GROUP_LABEL,
            cmdnames.MOUSE_REVIEW_TOGGLE,
        ),
        KeyboardCommand(
            "routePointerToItemHandler",
            owner.route_pointer_to_item,
            owner.GROUP_LABEL,
            cmdnames.ROUTE_POINTER_TO_ITEM,
            desktop_keybinding=kb_orca_kp_divide,
            laptop_keybinding=kb_orca_9,
        ),
        KeyboardCommand(
            "leftClickReviewItemHandler",
            owner.left_click_item,
            owner.GROUP_LABEL,
            cmdnames.LEFT_CLICK_REVIEW_ITEM,
            desktop_keybinding=kb_unmodified_kp_divide,
            laptop_keybinding=kb_orca_7,
        ),
        KeyboardCommand(
            "rightClickReviewItemHandler",
            owner.right_click_item,
            owner.GROUP_LABEL,
            cmdnames.RIGHT_CLICK_REVIEW_ITEM,
            desktop_keybinding=kb_unmodified_kp_multiply,
            laptop_keybinding=kb_orca_8,
        ),
    ]



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2011-2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for structural navigation."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from collections.abc import Callable

    from .input_event import InputEvent
    from .scripts import default
    from .structural_navigator import StructuralNavigator

    CommandCallback = Callable[[default.Script, InputEvent | None, bool], bool]


NAVIGATION_BINDINGS = (
    ("q", "blockquote", "blockquotes"),
    ("b", "button", "buttons"),
    ("x", "checkbox", "checkboxes"),
    ("c", "combobox", "comboboxes"),
    ("e", "entry", "entries"),
    ("f", "form_field", "form_fields"),
    ("h", "heading", "headings"),
    ("g", "image", "images"),
    ("m", "landmark", "landmarks"),
    ("l", "list", "lists"),
    ("i", "list_item", "list_items"),
    ("p", "paragraph", "paragraphs"),
    ("r", "radio_button", "radio_buttons"),
    ("t", "table", "tables"),
    ("k", "link", "links"),
    ("u", "unvisited_link", "unvisited_links"),
    ("v", "visited_link", "visited_links"),
    ("o", "large_object", "large_objects"),
    ("a", "clickable", "clickables"),
)

COMMAND_SPECS = (
    ("previous_annotation", "previous_annotation", cmdnames.ANNOTATION_PREV),
    ("next_annotation", "next_annotation", cmdnames.ANNOTATION_NEXT),
    ("list_annotations", "list_annotations", cmdnames.ANNOTATION_LIST),
    ("previous_blockquote", "previous_blockquote", cmdnames.BLOCKQUOTE_PREV),
    ("next_blockquote", "next_blockquote", cmdnames.BLOCKQUOTE_NEXT),
    ("list_blockquotes", "list_blockquotes", cmdnames.BLOCKQUOTE_LIST),
    ("previous_button", "previous_button", cmdnames.BUTTON_PREV),
    ("next_button", "next_button", cmdnames.BUTTON_NEXT),
    ("list_buttons", "list_buttons", cmdnames.BUTTON_LIST),
    ("previous_checkbox", "previous_checkbox", cmdnames.CHECK_BOX_PREV),
    ("next_checkbox", "next_checkbox", cmdnames.CHECK_BOX_NEXT),
    ("list_checkboxes", "list_checkboxes", cmdnames.CHECK_BOX_LIST),
    ("previous_combobox", "previous_combobox", cmdnames.COMBO_BOX_PREV),
    ("next_combobox", "next_combobox", cmdnames.COMBO_BOX_NEXT),
    ("list_comboboxes", "list_comboboxes", cmdnames.COMBO_BOX_LIST),
    ("previous_entry", "previous_entry", cmdnames.ENTRY_PREV),
    ("next_entry", "next_entry", cmdnames.ENTRY_NEXT),
    ("list_entries", "list_entries", cmdnames.ENTRY_LIST),
    ("previous_form_field", "previous_form_field", cmdnames.FORM_FIELD_PREV),
    ("next_form_field", "next_form_field", cmdnames.FORM_FIELD_NEXT),
    ("list_form_fields", "list_form_fields", cmdnames.FORM_FIELD_LIST),
    ("previous_heading", "previous_heading", cmdnames.HEADING_PREV),
    ("next_heading", "next_heading", cmdnames.HEADING_NEXT),
    ("list_headings", "list_headings", cmdnames.HEADING_LIST),
    ("previous_iframe", "previous_iframe", cmdnames.IFRAME_PREV),
    ("next_iframe", "next_iframe", cmdnames.IFRAME_NEXT),
    ("list_iframes", "list_iframes", cmdnames.IFRAME_LIST),
    ("previous_image", "previous_image", cmdnames.IMAGE_PREV),
    ("next_image", "next_image", cmdnames.IMAGE_NEXT),
    ("list_images", "list_images", cmdnames.IMAGE_LIST),
    ("previous_landmark", "previous_landmark", cmdnames.LANDMARK_PREV),
    ("next_landmark", "next_landmark", cmdnames.LANDMARK_NEXT),
    ("list_landmarks", "list_landmarks", cmdnames.LANDMARK_LIST),
    ("previous_list", "previous_list", cmdnames.LIST_PREV),
    ("next_list", "next_list", cmdnames.LIST_NEXT),
    ("list_lists", "list_lists", cmdnames.LIST_LIST),
    ("previous_list_item", "previous_list_item", cmdnames.LIST_ITEM_PREV),
    ("next_list_item", "next_list_item", cmdnames.LIST_ITEM_NEXT),
    ("list_list_items", "list_list_items", cmdnames.LIST_ITEM_LIST),
    ("previous_live_region", "previous_live_region", cmdnames.LIVE_REGION_PREV),
    ("next_live_region", "next_live_region", cmdnames.LIVE_REGION_NEXT),
    ("last_live_region", "_last_live_region", cmdnames.LIVE_REGION_LAST),
    ("previous_math", "previous_math", cmdnames.MATH_PREV),
    ("next_math", "next_math", cmdnames.MATH_NEXT),
    ("list_math", "list_math", cmdnames.MATH_LIST),
    ("previous_paragraph", "previous_paragraph", cmdnames.PARAGRAPH_PREV),
    ("next_paragraph", "next_paragraph", cmdnames.PARAGRAPH_NEXT),
    ("list_paragraphs", "list_paragraphs", cmdnames.PARAGRAPH_LIST),
    ("previous_radio_button", "previous_radio_button", cmdnames.RADIO_BUTTON_PREV),
    ("next_radio_button", "next_radio_button", cmdnames.RADIO_BUTTON_NEXT),
    ("list_radio_buttons", "list_radio_buttons", cmdnames.RADIO_BUTTON_LIST),
    ("previous_separator", "previous_separator", cmdnames.SEPARATOR_PREV),
    ("next_separator", "next_separator", cmdnames.SEPARATOR_NEXT),
    ("previous_table", "previous_table", cmdnames.TABLE_PREV),
    ("next_table", "next_table", cmdnames.TABLE_NEXT),
    ("list_tables", "list_tables", cmdnames.TABLE_LIST),
    ("previous_link", "previous_link", cmdnames.LINK_PREV),
    ("next_link", "next_link", cmdnames.LINK_NEXT),
    ("list_links", "list_links", cmdnames.LINK_LIST),
    ("previous_unvisited_link", "previous_unvisited_link", cmdnames.UNVISITED_LINK_PREV),
    ("next_unvisited_link", "next_unvisited_link", cmdnames.UNVISITED_LINK_NEXT),
    ("list_unvisited_links", "list_unvisited_links", cmdnames.UNVISITED_LINK_LIST),
    ("previous_visited_link", "previous_visited_link", cmdnames.VISITED_LINK_PREV),
    ("next_visited_link", "next_visited_link", cmdnames.VISITED_LINK_NEXT),
    ("list_visited_links", "list_visited_links", cmdnames.VISITED_LINK_LIST),
    ("previous_large_object", "previous_large_object", cmdnames.LARGE_OBJECT_PREV),
    ("next_large_object", "next_large_object", cmdnames.LARGE_OBJECT_NEXT),
    ("list_large_objects", "list_large_objects", cmdnames.LARGE_OBJECT_LIST),
    ("previous_clickable", "previous_clickable", cmdnames.CLICKABLE_PREV),
    ("next_clickable", "next_clickable", cmdnames.CLICKABLE_NEXT),
    ("list_clickables", "list_clickables", cmdnames.CLICKABLE_LIST),
    ("container_start", "container_start", cmdnames.CONTAINER_START),
    ("container_end", "container_end", cmdnames.CONTAINER_END),
)


def _get_command_bindings() -> dict[str, keybindings.KeyBinding | None]:
    """Returns command name to keybinding mapping."""

    command_bindings: dict[str, keybindings.KeyBinding | None] = {}
    for key, singular, plural in NAVIGATION_BINDINGS:
        command_bindings[f"previous_{singular}"] = keybindings.KeyBinding(
            key,
            keybindings.SHIFT_MODIFIER_MASK,
        )
        command_bindings[f"next_{singular}"] = keybindings.KeyBinding(
            key,
            keybindings.NO_MODIFIER_MASK,
        )
        command_bindings[f"list_{plural}"] = keybindings.KeyBinding(
            key,
            keybindings.SHIFT_ALT_MODIFIER_MASK,
        )

    command_bindings["previous_separator"] = keybindings.KeyBinding(
        "s",
        keybindings.SHIFT_MODIFIER_MASK,
    )
    command_bindings["next_separator"] = keybindings.KeyBinding(
        "s",
        keybindings.NO_MODIFIER_MASK,
    )
    command_bindings["previous_live_region"] = keybindings.KeyBinding(
        "d",
        keybindings.SHIFT_MODIFIER_MASK,
    )
    command_bindings["next_live_region"] = keybindings.KeyBinding(
        "d",
        keybindings.NO_MODIFIER_MASK,
    )
    command_bindings["last_live_region"] = keybindings.KeyBinding(
        "y",
        keybindings.NO_MODIFIER_MASK,
    )
    command_bindings["container_start"] = keybindings.KeyBinding(
        "comma",
        keybindings.SHIFT_MODIFIER_MASK,
    )
    command_bindings["container_end"] = keybindings.KeyBinding(
        "comma",
        keybindings.NO_MODIFIER_MASK,
    )
    command_bindings["previous_annotation"] = None
    command_bindings["next_annotation"] = None
    command_bindings["list_annotations"] = None
    command_bindings["previous_iframe"] = None
    command_bindings["next_iframe"] = None
    command_bindings["list_iframes"] = None
    command_bindings["previous_math"] = None
    command_bindings["next_math"] = None
    command_bindings["list_math"] = None
    return command_bindings


def _command_callback(owner: StructuralNavigator, method_name: str) -> CommandCallback:
    """Returns the command callback for method_name."""

    return getattr(owner, method_name)  # pylint: disable=protected-access


def get_commands(owner: StructuralNavigator) -> list[Command]:
    """Returns commands for structural navigation."""

    kb_orca_z = keybindings.KeyBinding("z", keybindings.ORCA_MODIFIER_MASK)
    commands: list[Command] = [
        KeyboardCommand(
            "structural_navigator_mode_cycle",
            owner.cycle_mode,
            owner.GROUP_LABEL,
            cmdnames.STRUCTURAL_NAVIGATION_MODE_CYCLE,
            desktop_keybinding=kb_orca_z,
            laptop_keybinding=kb_orca_z,
            is_group_toggle=True,
        ),
    ]

    command_bindings = _get_command_bindings()
    for name, method_name, description in COMMAND_SPECS:
        kb = command_bindings.get(name)
        commands.append(
            KeyboardCommand(
                name,
                _command_callback(owner, method_name),
                owner.GROUP_LABEL,
                description,
                desktop_keybinding=kb,
                laptop_keybinding=kb,
            ),
        )

    for level in range(1, 7):
        kb_previous = keybindings.KeyBinding(str(level), keybindings.SHIFT_MODIFIER_MASK)
        kb_next = keybindings.KeyBinding(str(level), keybindings.NO_MODIFIER_MASK)
        kb_list = keybindings.KeyBinding(str(level), keybindings.SHIFT_ALT_MODIFIER_MASK)

        heading_commands = (
            (
                f"previous_heading_level_{level}",
                f"previous_heading_level_{level}",
                cmdnames.HEADING_AT_LEVEL_PREV % level,
                kb_previous,
            ),
            (
                f"next_heading_level_{level}",
                f"next_heading_level_{level}",
                cmdnames.HEADING_AT_LEVEL_NEXT % level,
                kb_next,
            ),
            (
                f"list_headings_level_{level}",
                f"list_headings_level_{level}",
                cmdnames.HEADING_AT_LEVEL_LIST % level,
                kb_list,
            ),
        )
        for name, method_name, description, kb in heading_commands:
            commands.append(
                KeyboardCommand(
                    name,
                    _command_callback(owner, method_name),
                    owner.GROUP_LABEL,
                    description,
                    desktop_keybinding=kb,
                    laptop_keybinding=kb,
                ),
            )

    return commands



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Manager for text attribute presentation preferences."""

from __future__ import annotations

import enum
from typing import TYPE_CHECKING, Any

from . import dbus_service, debug, gsettings_registry
from .ax_text import AXTextAttribute

if TYPE_CHECKING:
    from .text_attribute_manager_preferences_grid import TextAttributePreferencesGrid


class PresentationMode(enum.IntEnum):
    """How a text attribute should be presented."""

    NONE = 0
    SPEAK = 1
    BRAILLE = 2
    SPEAK_AND_BRAILLE = 3


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.TextAttributeChangeMode",
    values={"off": 0, "editable-only": 1, "always": 2},
)
class TextAttributeChangeMode(enum.Enum):
    """When to announce text attribute changes during navigation."""

    OFF = 0
    EDITABLE_ONLY = 1
    ALWAYS = 2


@gsettings_registry.get_registry().gsettings_schema(
    "org.gnome.Orca.TextAttributes",
    name="text-attributes",
)
class TextAttributeManager:
    """Manager for text attribute presentation settings."""

    _SCHEMA = "text-attributes"
    KEY_ATTRIBUTES_TO_SPEAK = "attributes-to-speak"
    KEY_ATTRIBUTES_TO_BRAILLE = "attributes-to-braille"

    def _get_setting(self, key: str, gtype: str, default: Any) -> Any:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            key,
            gtype,
            default=default,
        )

    def __init__(self) -> None:
        msg = "TEXT ATTRIBUTE MANAGER: Registering D-Bus commands."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        controller = dbus_service.get_remote_controller()
        controller.register_decorated_module("TextAttributeManager", self)

    def create_preferences_grid(self) -> TextAttributePreferencesGrid:
        """Create and return a preferences grid for text attributes."""

        # pylint: disable-next=import-outside-toplevel
        from .text_attribute_manager_preferences_grid import TextAttributePreferencesGrid

        return TextAttributePreferencesGrid()

    @gsettings_registry.get_registry().gsetting(
        key=KEY_ATTRIBUTES_TO_SPEAK,
        schema="text-attributes",
        gtype="as",
        default=[],
        summary="Text attributes to speak",
        migration_key="textAttributesToSpeak",
    )
    @dbus_service.getter
    def get_attributes_to_speak(self) -> list[str]:
        """Returns the list of text attributes to speak."""

        return self._get_setting(self.KEY_ATTRIBUTES_TO_SPEAK, "as", [])

    def get_resolved_attributes_to_speak(self) -> list[AXTextAttribute]:
        """Returns the resolved list of attributes to speak, falling back to defaults."""

        result = list(
            filter(None, map(AXTextAttribute.from_string, self.get_attributes_to_speak()))
        )
        if not result:
            result = [a for a in AXTextAttribute if a.should_present_by_default()]
        return result

    @dbus_service.setter
    def set_attributes_to_speak(self, value: list[str]) -> bool:
        """Sets the list of text attributes to speak."""

        if self.get_attributes_to_speak() == value:
            return True

        msg = f"TEXT ATTRIBUTE MANAGER: Setting attributes to speak to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_ATTRIBUTES_TO_SPEAK,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_ATTRIBUTES_TO_BRAILLE,
        schema="text-attributes",
        gtype="as",
        default=[],
        summary="Text attributes to mark in braille",
        migration_key="textAttributesToBraille",
    )
    @dbus_service.getter
    def get_attributes_to_braille(self) -> list[str]:
        """Returns the list of text attributes to mark in braille."""

        return self._get_setting(self.KEY_ATTRIBUTES_TO_BRAILLE, "as", [])

    @dbus_service.setter
    def set_attributes_to_braille(self, value: list[str]) -> bool:
        """Sets the list of text attributes to mark in braille."""

        if self.get_attributes_to_braille() == value:
            return True

        msg = f"TEXT ATTRIBUTE MANAGER: Setting attributes to braille to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_ATTRIBUTES_TO_BRAILLE,
            value,
        )
        return True


_manager = TextAttributeManager()


def get_manager() -> TextAttributeManager:
    """Return the singleton TextAttributeManager instance."""

    return _manager



# Orca
#
# Copyright 2024-2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grid for sleep mode settings."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import gsettings_registry, guilabels, orca_gui_helpers, preferences_grid_base

if TYPE_CHECKING:
    import gi

    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk  # pylint: disable=no-name-in-module

    from .sleep_mode_manager import SleepModeManager


class SleepModePreferencesGrid(preferences_grid_base.PreferencesGridBase):
    """Preferences grid for sleep mode settings."""

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for sleep mode preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.KB_GROUP_SLEEP_MODE,
            panel_id="manual.sleep-mode",
            summary="Use these settings to choose applications where Orca should automatically "
            "enter sleep mode.",
            description=guilabels.SLEEP_MODE_INFO,
            schema="sleep-mode",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SLEEP_MODE_APPS,
                    kind="list",
                    summary="Lists the applications for which you can enable automatic sleep mode.",
                    schema="sleep-mode",
                    key="apps",
                    dynamic_values=True,
                ),
            ),
        )

    def __init__(self, manager: SleepModeManager) -> None:
        super().__init__(guilabels.KB_GROUP_SLEEP_MODE)
        self._manager: SleepModeManager = manager
        self._app_switches: list[tuple[str, Gtk.Switch]] = []
        self._listbox: preferences_grid_base.PreferencesFocusManagedListBox | None = None
        self._build()
        self.refresh()

    def _build(self) -> None:
        """Create the Gtk widgets composing the grid."""

        row = 0

        info_listbox = orca_gui_helpers.create_info_listbox(guilabels.SLEEP_MODE_INFO)
        self.attach(info_listbox, 0, row, 1, 1)
        row += 1

        section_box, section_content, _heading_label = orca_gui_helpers.create_section_box(
            guilabels.SLEEP_MODE_APPS,
            margin_top=0,
        )
        self._scrolled = orca_gui_helpers.create_preferences_scrolled_window()
        orca_gui_helpers.add_section_content(section_box, section_content, self._scrolled)
        self.attach(section_box, 0, row, 1, 1)
        self._scrolled_row = row

    def _create_app_row(self, app_name: str, is_enabled: bool) -> None:
        """Create a switch row for an app and add it to the listbox."""

        assert self._listbox is not None
        row, switch, _label = orca_gui_helpers.create_switch_row(
            app_name,
            self._on_switch_toggled,
            is_enabled,
            include_top_separator=False,
        )
        self._listbox.add_row_with_widget(row, switch)
        self._app_switches.append((app_name, switch))

    def _on_switch_toggled(self, _switch: Gtk.Switch, _pspec: object) -> None:
        """Handle switch toggle."""

        self._has_unsaved_changes = True

    def reload(self) -> None:
        """Reload settings and refresh the UI."""

        self._has_unsaved_changes = False
        self.refresh()

    def refresh(self) -> None:
        """Rebuild the app list with a fresh PreferencesFocusManagedListBox."""

        # pylint: disable=no-member
        self._app_switches.clear()
        self._listbox = preferences_grid_base.PreferencesFocusManagedListBox()
        self._listbox.get_accessible().set_name(guilabels.SLEEP_MODE_APPS)

        saved_apps = set(self._manager.get_sleep_mode_apps())
        running_apps = set(self._manager.get_running_app_names())
        all_apps = running_apps | saved_apps

        for name in sorted(all_apps, key=lambda n: (n not in saved_apps, n.lower())):
            self._create_app_row(name, is_enabled=name in saved_apps)

        child = self._scrolled.get_child()
        if child is not None:
            self._scrolled.remove(child)
        self._scrolled.add(self._listbox)
        self._scrolled.show_all()
        # pylint: enable=no-member

    def save_settings(self, profile: str = "", app_name: str = "") -> dict[str, list[str]]:
        """Save the list of sleep-mode apps to dconf."""

        enabled_apps = [name for name, switch in self._app_switches if switch.get_active()]
        self._manager.set_sleep_mode_apps(enabled_apps)
        self._has_unsaved_changes = False

        if profile:
            registry = gsettings_registry.get_registry()
            registry.save_schema(
                "sleep-mode",
                {"apps": enabled_apps},
                profile,
                app_name,
            )

        return {"apps": enabled_apps}



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Utilities for reconstructing MathML from the accessibility tree."""

from __future__ import annotations

from html import escape
from typing import TYPE_CHECKING, ClassVar

from . import debug
from .ax_object import AXObject
from .ax_utilities_role import AXUtilitiesRole

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi


class AXUtilitiesMath:
    """Utilities for reconstructing MathML from accessible math objects."""

    _MATHML_ATTRIBUTES: ClassVar[dict[str, list[str]]] = {
        "menclose": ["notation"],
        "mfenced": ["open", "close", "separators"],
        "mfrac": ["linethickness"],
        "mo": ["form", "stretchy", "fence", "separator", "largeop", "movablelimits", "accent"],
        "mspace": ["width", "height", "depth"],
        "mstyle": ["mathvariant", "mathsize", "mathcolor", "mathbackground"],
        "mpadded": ["width", "height", "depth", "lspace", "voffset"],
        "mtable": ["align", "columnalign", "rowalign"],
        "mtr": ["columnalign", "rowalign"],
        "mtd": ["columnalign", "rowalign", "columnspan", "rowspan"],
        "mi": ["mathvariant"],
        "mn": ["mathvariant"],
        "ms": ["lquote", "rquote"],
    }

    _TOKEN_TAGS: ClassVar[set[str]] = {"mi", "mn", "mo", "mtext", "ms", "mspace"}

    @staticmethod
    def get_mathml(obj: Atspi.Accessible) -> str:
        """Returns a MathML string reconstructed from the accessible tree rooted at obj."""

        math_root = AXUtilitiesMath.find_math_root(obj)
        if math_root is None:
            return ""

        parts: list[str] = []
        AXUtilitiesMath._serialize(math_root, parts)
        result = "".join(parts)
        msg = f"AXUtilitiesMath: Reconstructed MathML: {result}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    @staticmethod
    def find_math_root(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Walks up the tree to find the top-level math element."""

        if not AXObject.is_valid(obj):
            return None

        if AXUtilitiesRole.is_math(obj):
            return obj

        candidate = obj
        while AXObject.is_valid(candidate):
            if AXUtilitiesRole.is_math(candidate):
                return candidate
            parent = AXObject.get_parent(candidate)
            if parent is None or parent == candidate:
                break
            candidate = parent

        if AXUtilitiesRole.is_math_related(obj):
            tokens = ["AXUtilitiesMath: Could not find math root from", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return None

    @staticmethod
    def _get_tag(obj: Atspi.Accessible) -> str:
        """Returns the MathML tag for an accessible object."""

        return AXObject.get_attribute(obj, "tag")

    @staticmethod
    def _get_text_content(obj: Atspi.Accessible) -> str:
        """Returns the text content for a MathML token element."""

        name = AXObject.get_name(obj)
        if name:
            return name

        # Some token elements might have children that contain the text.
        parts: list[str] = []
        for child in AXObject.iter_children(obj):
            child_name = AXObject.get_name(child)
            if child_name:
                parts.append(child_name)
        return "".join(parts)

    @staticmethod
    def _serialize(obj: Atspi.Accessible, parts: list[str]) -> None:
        """Recursively serializes an accessible math object into MathML parts."""

        tag = AXUtilitiesMath._get_tag(obj)
        if not tag:
            tokens = ["AXUtilitiesMath: No tag found for", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            text = AXObject.get_name(obj)
            if text:
                parts.append(escape(text))
            return

        # Build the opening tag with attributes.
        parts.append(f"<{tag}")
        attrs = AXObject.get_attributes_dict(obj)
        attr_names = AXUtilitiesMath._MATHML_ATTRIBUTES.get(tag, [])
        for attr_name in attr_names:
            value = attrs.get(attr_name)
            if value:
                parts.append(f' {attr_name}="{escape(value, quote=True)}"')
        parts.append(">")

        if tag in AXUtilitiesMath._TOKEN_TAGS:
            text = AXUtilitiesMath._get_text_content(obj)
            parts.append(escape(text))
        else:
            for child in AXObject.iter_children(obj):
                AXUtilitiesMath._serialize(child, parts)

        parts.append(f"</{tag}>")



# Orca
#
# Copyright 2006-2008 Sun Microsystems Inc.
# Copyright 2022 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for flat-review find."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .flat_review_finder import FlatReviewFinder


def get_commands(owner: FlatReviewFinder) -> list[Command]:
    """Returns commands for flat-review find."""

    kb_unmodified_kp_delete = keybindings.KeyBinding(
        "KP_Delete",
        keybindings.NO_MODIFIER_MASK,
    )
    kb_orca_bracketleft = keybindings.KeyBinding(
        "bracketleft",
        keybindings.ORCA_MODIFIER_MASK,
    )
    kb_orca_kp_delete = keybindings.KeyBinding(
        "KP_Delete",
        keybindings.ORCA_MODIFIER_MASK,
    )
    kb_orca_bracketright = keybindings.KeyBinding(
        "bracketright",
        keybindings.ORCA_MODIFIER_MASK,
    )
    kb_orca_shift_kp_delete = keybindings.KeyBinding(
        "KP_Delete",
        keybindings.ORCA_SHIFT_MODIFIER_MASK,
    )
    kb_orca_ctrl_bracketright = keybindings.KeyBinding(
        "bracketright",
        keybindings.ORCA_CTRL_MODIFIER_MASK,
    )

    return [
        KeyboardCommand(
            "findHandler",
            owner.show_dialog,
            owner.GROUP_LABEL,
            cmdnames.SHOW_FIND_GUI,
            desktop_keybinding=kb_unmodified_kp_delete,
            laptop_keybinding=kb_orca_bracketleft,
        ),
        KeyboardCommand(
            "findNextHandler",
            owner.find_next,
            owner.GROUP_LABEL,
            cmdnames.FIND_NEXT,
            desktop_keybinding=kb_orca_kp_delete,
            laptop_keybinding=kb_orca_bracketright,
        ),
        KeyboardCommand(
            "findPreviousHandler",
            owner.find_previous,
            owner.GROUP_LABEL,
            cmdnames.FIND_PREVIOUS,
            desktop_keybinding=kb_orca_shift_kp_delete,
            laptop_keybinding=kb_orca_ctrl_bracketright,
        ),
    ]



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods

"""Provides MathCAT-based math presentation support for speech and braille."""

from __future__ import annotations

import contextlib
import os
from enum import Enum
from typing import TYPE_CHECKING

from . import (  # pylint: disable=no-name-in-module
    dbus_service,
    debug,
    gsettings_registry,
    language_utilities,
    orca_platform,
)
from .ax_utilities_math import AXUtilitiesMath
from .extension import Extension

try:
    from . import libmathcat_py  # type: ignore[attr-defined]

    _LIB_AVAILABLE = True
except ImportError:
    libmathcat_py = None  # type: ignore[assignment]
    _LIB_AVAILABLE = False

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .math_presenter_preferences_grid import MathPreferencesGrid


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.MathCopyFormat",
    values={"mathml": 0, "latex": 1, "asciimath": 2, "speech": 3},
)
class MathCopyFormat(Enum):
    """Format for copying math content to the clipboard."""

    MATHML = 0
    LATEX = 1
    ASCIIMATH = 2
    SPEECH = 3


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.MathNavMode",
    values={"enhanced": 0, "simple": 1, "character": 2},
)
class MathNavMode(Enum):
    """Navigation granularity for math exploration."""

    ENHANCED = 0
    SIMPLE = 1
    CHARACTER = 2


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.MathBrailleNavHighlight",
    values={"off": 0, "first-char": 1, "end-points": 2, "all": 3},
)
class MathBrailleNavHighlight(Enum):
    """Braille navigation highlight style for the current math node."""

    OFF = 0
    FIRST_CHAR = 1
    END_POINTS = 2
    ALL = 3


@gsettings_registry.get_registry().gsettings_schema(
    "org.gnome.Orca.MathPresentation", name="math-presentation"
)
class MathPresenter(Extension):
    """Provides MathCAT-based math presentation support."""

    _SCHEMA = "math-presentation"

    KEY_LANGUAGE = "language"
    KEY_SPEECH_STYLE = "speech-style"
    KEY_VERBOSITY = "verbosity"
    KEY_BRAILLE_CODE = "braille-code"
    KEY_COPY_FORMAT = "copy-format"
    KEY_NAV_MODE = "nav-mode"
    KEY_BRAILLE_NAV_HIGHLIGHT = "braille-nav-highlight"
    KEY_AUTO_ZOOM_OUT = "auto-zoom-out"

    def __init__(self) -> None:
        self._initialized = False
        self._available = False
        super().__init__()

    def _get_setting(self, key: str, gtype: str, default: str) -> str:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            key,
            gtype,
            default=default,
        )

    def _find_rules_dir(self) -> str:
        """Returns the path to the MathCAT Rules directory, or empty string if not found."""

        if configured_dir := orca_platform.mathcat_rules_dir:
            if os.path.isdir(configured_dir):
                return configured_dir

        return ""

    def initialize(self) -> None:
        """Initializes MathCAT. Called lazily on first use."""

        if self._initialized:
            return

        self._initialized = True

        if not _LIB_AVAILABLE:
            msg = "MATH PRESENTER: libmathcat_py module not found"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        if not (rules_dir := self._find_rules_dir()):
            msg = "MATH PRESENTER: No Rules directory found"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        try:
            libmathcat_py.SetRulesDir(rules_dir)
        except OSError as err:
            msg = f"MATH PRESENTER: SetRulesDir failed for {rules_dir}: {err}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        version = libmathcat_py.GetVersion()
        msg = f"MATH PRESENTER: Initialized version {version} with rules from {rules_dir}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        libmathcat_py.SetPreference("TTS", "none")
        self._available = True
        self._apply_preferences()

    def is_available(self) -> bool:
        """Returns True if MathCAT is loaded and initialized."""

        self.initialize()
        return self._available

    def _resolve_language(self) -> str:
        """Returns the language code to use, resolving 'Auto' to system locale."""

        lang = self.get_language()
        if lang and lang != "Auto":
            return lang

        if language := language_utilities.get_current_language():
            return language

        return "en"

    def _apply_preferences(self) -> None:
        """Applies stored preferences to MathCAT."""

        if lang_code := self._resolve_language():
            try:
                supported = libmathcat_py.GetSupportedLanguages()
            except OSError as err:
                msg = f"MATH PRESENTER: GetSupportedLanguages failed: {err}"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                supported = []

            if lang_code not in supported:
                msg = f"MATH PRESENTER: Language {lang_code} not supported, "
                if "en" in supported:
                    lang_code = "en"
                elif supported:
                    lang_code = supported[0]
                else:
                    lang_code = ""
                msg += f"falling back to {lang_code!r}"
                debug.print_message(debug.LEVEL_INFO, msg, True)

            if lang_code:
                libmathcat_py.SetPreference("Language", lang_code)
                msg = f"MATH PRESENTER: Language set to {lang_code}"
                debug.print_message(debug.LEVEL_INFO, msg, True)

        if speech_style := self.get_speech_style():
            libmathcat_py.SetPreference("SpeechStyle", speech_style)

        if verbosity := self.get_verbosity():
            libmathcat_py.SetPreference("Verbosity", verbosity)

        if braille_code := self.get_braille_code():
            libmathcat_py.SetPreference("BrailleCode", braille_code)

        if nav_mode := self.get_nav_mode():
            libmathcat_py.SetPreference("NavMode", nav_mode.capitalize())

        if braille_highlight := self.get_braille_nav_highlight():
            nick_to_mathcat = {
                "off": "Off",
                "first-char": "FirstChar",
                "end-points": "EndPoints",
                "all": "All",
            }
            libmathcat_py.SetPreference(
                "BrailleNavHighlight",
                nick_to_mathcat.get(braille_highlight, "Off"),
            )

        libmathcat_py.SetPreference("AutoZoomOut", str(self.get_auto_zoom_out()).lower())

    @gsettings_registry.get_registry().gsetting(
        key=KEY_LANGUAGE,
        schema="math-presentation",
        gtype="s",
        default="Auto",
        summary="Math speech language (Auto uses system locale)",
        migration_key="mathLanguage",
    )
    @dbus_service.getter
    def get_language(self) -> str:
        """Returns the current math language setting."""

        return self._get_setting(self.KEY_LANGUAGE, "s", "Auto")

    @dbus_service.setter
    def set_language(self, value: str) -> bool:
        """Sets the math language."""

        msg = f"MATH PRESENTER: Setting language to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_LANGUAGE, value, "s"
        )
        if self._available:
            lang_code = value if value != "Auto" else language_utilities.get_current_language()
            libmathcat_py.SetPreference("Language", lang_code)
        return True

    def get_language_choices(self) -> list[str]:
        """Returns available languages, with 'Auto' as the first option."""

        choices = ["Auto"]
        if not self.is_available():
            return choices

        with contextlib.suppress(OSError):
            choices.extend(libmathcat_py.GetSupportedLanguages())
        return choices

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SPEECH_STYLE,
        schema="math-presentation",
        gtype="s",
        default="ClearSpeak",
        summary="Math speech style",
        migration_key="mathSpeechStyle",
    )
    @dbus_service.getter
    def get_speech_style(self) -> str:
        """Returns the current math speech style."""

        return self._get_setting(self.KEY_SPEECH_STYLE, "s", "ClearSpeak")

    @dbus_service.setter
    def set_speech_style(self, value: str) -> bool:
        """Sets the math speech style."""

        msg = f"MATH PRESENTER: Setting speech style to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_SPEECH_STYLE, value, "s"
        )
        if self._available:
            libmathcat_py.SetPreference("SpeechStyle", value)
        return True

    def get_speech_style_choices(self) -> list[str]:
        """Returns available speech styles."""

        if not self.is_available():
            return ["ClearSpeak", "SimpleSpeak"]

        try:
            lang = libmathcat_py.GetPreference("Language")
            return libmathcat_py.GetSupportedSpeechStyles(lang)
        except OSError:
            return ["ClearSpeak", "SimpleSpeak"]

    @gsettings_registry.get_registry().gsetting(
        key=KEY_VERBOSITY,
        schema="math-presentation",
        gtype="s",
        default="Medium",
        summary="Math speech verbosity",
        migration_key="mathVerbosity",
    )
    @dbus_service.getter
    def get_verbosity(self) -> str:
        """Returns the current math speech verbosity."""

        return self._get_setting(self.KEY_VERBOSITY, "s", "Medium")

    @dbus_service.setter
    def set_verbosity(self, value: str) -> bool:
        """Sets the math speech verbosity."""

        msg = f"MATH PRESENTER: Setting verbosity to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_VERBOSITY, value, "s"
        )
        if self._available:
            libmathcat_py.SetPreference("Verbosity", value)
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_BRAILLE_CODE,
        schema="math-presentation",
        gtype="s",
        default="Nemeth",
        summary="Math braille code",
        migration_key="mathBrailleCode",
    )
    @dbus_service.getter
    def get_braille_code(self) -> str:
        """Returns the current math braille code."""

        return self._get_setting(self.KEY_BRAILLE_CODE, "s", "Nemeth")

    @dbus_service.setter
    def set_braille_code(self, value: str) -> bool:
        """Sets the math braille code."""

        msg = f"MATH PRESENTER: Setting braille code to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_BRAILLE_CODE, value, "s"
        )
        if self._available:
            libmathcat_py.SetPreference("BrailleCode", value)
        return True

    def get_braille_code_choices(self) -> list[str]:
        """Returns available braille codes."""

        if not self.is_available():
            return ["Nemeth", "UEB"]

        try:
            return libmathcat_py.GetSupportedBrailleCodes()
        except OSError:
            return ["Nemeth", "UEB"]

    @gsettings_registry.get_registry().gsetting(
        key=KEY_COPY_FORMAT,
        schema="math-presentation",
        genum="org.gnome.Orca.MathCopyFormat",
        default="mathml",
        summary="Format for copying math content to clipboard",
        migration_key="mathCopyFormat",
    )
    @dbus_service.getter
    def get_copy_format(self) -> str:
        """Returns the format used when copying math content."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_COPY_FORMAT,
            "",
            genum="org.gnome.Orca.MathCopyFormat",
            default="mathml",
        )

    @dbus_service.setter
    def set_copy_format(self, value: str) -> bool:
        """Sets the format used when copying math content."""

        msg = f"MATH PRESENTER: Setting copy format to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_COPY_FORMAT,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_NAV_MODE,
        schema="math-presentation",
        genum="org.gnome.Orca.MathNavMode",
        default="enhanced",
        summary="Math navigation granularity (enhanced, simple, character)",
        migration_key="mathNavMode",
    )
    @dbus_service.getter
    def get_nav_mode(self) -> str:
        """Returns the math navigation mode."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_NAV_MODE,
            "",
            genum="org.gnome.Orca.MathNavMode",
            default="enhanced",
        )

    @dbus_service.setter
    def set_nav_mode(self, value: str) -> bool:
        """Sets the math navigation mode."""

        msg = f"MATH PRESENTER: Setting nav mode to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_NAV_MODE,
            value,
        )
        if self._available:
            mathcat_value = value.capitalize()
            libmathcat_py.SetPreference("NavMode", mathcat_value)
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_BRAILLE_NAV_HIGHLIGHT,
        schema="math-presentation",
        genum="org.gnome.Orca.MathBrailleNavHighlight",
        default="off",
        summary="Braille navigation highlight style (off, first-char, end-points, all)",
        migration_key="mathBrailleNavHighlight",
    )
    @dbus_service.getter
    def get_braille_nav_highlight(self) -> str:
        """Returns the braille navigation highlight style."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_BRAILLE_NAV_HIGHLIGHT,
            "",
            genum="org.gnome.Orca.MathBrailleNavHighlight",
            default="off",
        )

    @dbus_service.setter
    def set_braille_nav_highlight(self, value: str) -> bool:
        """Sets the braille navigation highlight style."""

        msg = f"MATH PRESENTER: Setting braille nav highlight to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_BRAILLE_NAV_HIGHLIGHT,
            value,
        )
        if self._available:
            nick_to_mathcat = {
                "off": "Off",
                "first-char": "FirstChar",
                "end-points": "EndPoints",
                "all": "All",
            }
            libmathcat_py.SetPreference("BrailleNavHighlight", nick_to_mathcat.get(value, "Off"))
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_AUTO_ZOOM_OUT,
        schema="math-presentation",
        gtype="b",
        default=False,
        summary="Automatically exit 2D math structures when moving past the edge",
        migration_key="mathAutoZoomOut",
    )
    @dbus_service.getter
    def get_auto_zoom_out(self) -> bool:
        """Returns whether auto zoom out is enabled."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_AUTO_ZOOM_OUT,
            "b",
            default=False,
        )

    @dbus_service.setter
    def set_auto_zoom_out(self, value: bool) -> bool:
        """Sets whether auto zoom out is enabled."""

        msg = f"MATH PRESENTER: Setting auto zoom out to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_AUTO_ZOOM_OUT,
            value,
        )
        if self._available:
            libmathcat_py.SetPreference("AutoZoomOut", str(value).lower())
        return True

    # pylint: disable-next=too-many-return-statements
    def get_navigation_content_for_copy(self) -> str:
        """Returns the current navigation node content in the configured copy format."""

        copy_format = self.get_copy_format()

        if copy_format == "mathml":
            try:
                mathml, _offset = libmathcat_py.GetNavigationMathML()
                return mathml.strip()
            except OSError as err:
                msg = f"MATH PRESENTER: GetNavigationMathML failed: {err}"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return ""

        if copy_format == "speech":
            try:
                return libmathcat_py.GetSpokenText()
            except OSError as err:
                msg = f"MATH PRESENTER: GetSpokenText failed: {err}"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return ""

        if copy_format in ("latex", "asciimath"):
            braille_code = "LaTeX" if copy_format == "latex" else "ASCIIMath"
            original_code = self.get_braille_code()
            try:
                libmathcat_py.SetPreference("BrailleCode", braille_code)
                return libmathcat_py.GetNavigationBraille()
            except OSError as err:
                msg = f"MATH PRESENTER: GetNavigationBraille({braille_code}) failed: {err}"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return ""
            finally:
                libmathcat_py.SetPreference("BrailleCode", original_code)

        return ""

    def create_preferences_grid(self) -> MathPreferencesGrid:
        """Returns the GtkGrid containing the math preferences UI."""

        # pylint: disable-next=import-outside-toplevel
        from .math_presenter_preferences_grid import MathPreferencesGrid

        return MathPreferencesGrid(self)

    def get_speech_for_math(self, obj: Atspi.Accessible) -> str:
        """Given a math accessible, returns MathCAT speech or empty string on failure."""

        if not self.is_available():
            return ""

        if not (mathml := AXUtilitiesMath.get_mathml(obj)):
            return ""

        try:
            libmathcat_py.SetMathML(mathml)
            speech = libmathcat_py.GetSpokenText()
        except OSError as err:
            msg = f"MATH PRESENTER: GetSpokenText failed: {err}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return ""

        lang = libmathcat_py.GetPreference("Language")
        style = libmathcat_py.GetPreference("SpeechStyle")
        verbosity = libmathcat_py.GetPreference("Verbosity")
        nav_mode = libmathcat_py.GetPreference("NavMode")
        msg = (
            f"MATH PRESENTER: Speech ({lang}, {style},"
            f" verbosity={verbosity}, nav={nav_mode}): {speech}"
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return speech

    def get_braille_for_math(self, obj: Atspi.Accessible) -> str:
        """Given a math accessible, returns MathCAT braille or empty string on failure."""

        if not self.is_available():
            return ""

        if not (mathml := AXUtilitiesMath.get_mathml(obj)):
            return ""

        try:
            libmathcat_py.SetMathML(mathml)
            braille = libmathcat_py.GetBraille("")
        except OSError as err:
            msg = f"MATH PRESENTER: GetBraille failed: {err}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return ""

        code = libmathcat_py.GetPreference("BrailleCode")
        highlight = libmathcat_py.GetPreference("BrailleNavHighlight")
        msg = f"MATH PRESENTER: Braille ({code}, highlight={highlight}): {braille}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return braille

    def get_where_am_i(self) -> str:
        """Returns MathCAT's where-am-I speech for the current navigation position."""

        try:
            return libmathcat_py.DoNavigateCommand("WhereAmI")
        except OSError as err:
            msg = f"MATH PRESENTER: WhereAmI failed: {err}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return ""

    def get_where_am_i_all(self) -> str:
        """Returns MathCAT's detailed where-am-I speech for the current navigation position."""

        try:
            return libmathcat_py.DoNavigateCommand("WhereAmIAll")
        except OSError as err:
            msg = f"MATH PRESENTER: WhereAmIAll failed: {err}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return ""


_presenter: MathPresenter = MathPresenter()


def get_presenter() -> MathPresenter:
    """Returns the Math Presenter"""

    return _presenter



# Orca
#
# Copyright 2025-2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-statements
# pylint: disable=too-many-locals
# pylint: disable=too-many-lines
# pylint: disable=too-many-instance-attributes

"""Preferences grid for Orca commands and keybindings."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, GLib, Gtk  # pylint: disable=no-name-in-module

from . import (
    ax_device_manager,
    command_manager,
    debug,
    gsettings_registry,
    guilabels,
    keybindings,
    keynames,
    messages,
    orca_gui_helpers,
    orca_modifier_manager,
    preferences_grid_base,
    presentation_manager,
    script_manager,
)
from .ax_object import AXObject

if TYPE_CHECKING:
    from collections.abc import Callable

    from .command import KeyboardCommand
    from .scripts import default


class CommandListBoxRow(Gtk.ListBoxRow):
    """ListBoxRow for command keybindings with command details."""

    def __init__(self) -> None:
        super().__init__()
        self.command: KeyboardCommand | None = None
        self.vbox: Gtk.Box | None = None
        self.binding_label: Gtk.Label | None = None


class KeybindingsPreferencesGrid(preferences_grid_base.PreferencesGridBase):
    """Grid widget for keybindings preferences."""

    # pylint: disable=no-member

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for commands preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.COMMANDS,
            panel_id="manual.commands",
            description=(
                "Commands settings let you choose the keyboard layout and screen reader "
                "modifier keys, and customize command keyboard shortcuts.\n\n"
                "Orca commands are organized into groups in the Commands list. Choose a "
                "group from the list, then activate its row, or press Right Arrow, to "
                "open its commands. Press Left Arrow, Escape, or Alt+Left to return to "
                "the list.\n\n"
                "The bindings shown here are for standard keyboards. Braille-display "
                "button bindings are not configured in the Commands list.\n\n"
                "After choosing a command, press the new key combination, then press "
                "Return to bind it. Press Delete or Backspace without modifiers to "
                "unbind the command. Press Escape to cancel. Orca warns you if the "
                "binding is already used by another active command."
            ),
            schema="keybindings",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.KEYBOARD_LAYOUT,
                    kind="choice",
                    summary="Choose the layout that matches your keyboard.",
                    schema="keybindings",
                    key="keyboard-layout",
                    values=(guilabels.KEYBOARD_LAYOUT_DESKTOP, guilabels.KEYBOARD_LAYOUT_LAPTOP),
                    value_docs=(
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.KEYBOARD_LAYOUT_DESKTOP,
                            value="desktop",
                            summary="Intended for use with a full-size keyboard. The desktop "
                            "layout uses the numeric keypad for some Orca commands.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.KEYBOARD_LAYOUT_LAPTOP,
                            value="laptop",
                            summary="Intended for use with a laptop-style keyboard, including "
                            "keyboards that lack a numeric keypad.",
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.MODIFIER_KEYS,
                    kind="list",
                    summary=(
                        "Chooses which keys act as the Orca modifier for the selected "
                        "keyboard layout."
                    ),
                    values=(
                        guilabels.MODIFIER_INSERT,
                        guilabels.MODIFIER_KP_INSERT,
                        guilabels.MODIFIER_CAPS_LOCK,
                    ),
                ),
            ),
        )

    def __init__(
        self,
        script: default.Script,
        title_change_callback: Callable[[str], None] | None = None,
    ) -> None:
        """Initialize the keybindings preferences grid."""

        super().__init__(guilabels.COMMANDS)
        self._script = script
        self._initializing = True
        self._title_change_callback = title_change_callback

        self._categories: dict[str, list[KeyboardCommand]] = {}
        self._current_category: str | None = None
        self._captured_key: tuple[str, int, int] = ("", 0, 0)
        self._orca_modifier_pressed_during_capture: bool = False
        # Store modified keybindings separately so they survive apply_user_overrides()
        self._modified_keybindings: dict[str, keybindings.KeyBinding | None] = {}
        self._keybinding_being_edited: str | None = None
        self._saved_commands: dict[str, KeyboardCommand] = {}

        self._original_keyboard_layout_is_desktop: bool = (
            command_manager.get_manager().get_keyboard_layout_is_desktop()
        )

        self.keyboard_layout_combo: Gtk.ComboBox | None = None
        self._modifier_switches: dict[str, Gtk.Switch] = {}
        self._settings_listbox: preferences_grid_base.PreferencesFocusManagedListBox | None = None
        self._commands_categories_listbox: Gtk.ListBox | None = None
        self._commands_detail_listbox: Gtk.ListBox | None = None

        self._build()
        self._initializing = False

    _MODIFIER_KEY_OPTIONS: ClassVar[list[tuple[str, list[str]]]] = [
        (guilabels.MODIFIER_INSERT, ["Insert"]),
        (guilabels.MODIFIER_KP_INSERT, ["KP_Insert"]),
        (guilabels.MODIFIER_CAPS_LOCK, ["Caps_Lock", "Shift_Lock"]),
    ]

    def _build(self) -> None:
        """Build the keybindings UI."""

        row = 0

        keyboard_layout_model = Gtk.ListStore(str, int)
        keyboard_layout_model.append(
            [guilabels.KEYBOARD_LAYOUT_DESKTOP, command_manager.KeyboardLayout.DESKTOP.value],
        )
        keyboard_layout_model.append(
            [guilabels.KEYBOARD_LAYOUT_LAPTOP, command_manager.KeyboardLayout.LAPTOP.value],
        )

        self._settings_listbox = preferences_grid_base.PreferencesFocusManagedListBox()

        layout_row, layout_combo, _label = orca_gui_helpers.create_combo_box_row(
            guilabels.KEYBOARD_LAYOUT,
            keyboard_layout_model,
            self._on_keyboard_layout_changed,
            include_top_separator=False,
        )
        self._settings_listbox.add_row_with_widget(layout_row, layout_combo)
        self.keyboard_layout_combo = layout_combo

        self.attach(self._settings_listbox, 0, row, 1, 1)
        row += 1

        modifier_listbox = preferences_grid_base.PreferencesFocusManagedListBox(
            guilabels.MODIFIER_KEYS
        )

        for label, keys in self._MODIFIER_KEY_OPTIONS:
            switch_row, switch, _label = orca_gui_helpers.create_switch_row(
                label,
                self._on_modifier_switch_toggled,
                state=False,
                include_top_separator=False,
            )
            self._modifier_switches[keys[0]] = switch
            modifier_listbox.add_row_with_widget(switch_row, switch)

        self.attach(modifier_listbox.get_container(), 0, row, 1, 1)
        row += 1

        section_box, section_content, commands_heading = orca_gui_helpers.create_section_box(
            guilabels.COMMANDS
        )
        stack, self._commands_categories_listbox, self._commands_detail_listbox = (
            self._create_list_detail_preferences_stack(
                on_category_activated=self._on_category_activated,
                on_detail_row_activated=self._on_keybinding_activated,
            )
        )
        self._commands_categories_listbox.get_accessible().set_name(guilabels.COMMANDS)
        orca_gui_helpers.add_section_content(section_box, section_content, stack)
        self.attach(section_box, 0, row, 1, 1)

        self._register_stack_disable_widgets(
            self._settings_listbox,
            modifier_listbox.get_container(),
            commands_heading,
        )

    def reload(self) -> None:
        """Reload keybindings from the script."""

        app_name = AXObject.get_name(self._script.app) if self._script.app else ""
        layout = gsettings_registry.get_registry().layered_lookup(
            "keybindings",
            "keyboard-layout",
            "",
            genum="org.gnome.Orca.KeyboardLayout",
            app_name=app_name or None,
            default="desktop",
        )
        command_manager.get_manager().set_keyboard_layout_is_desktop(layout == "desktop")
        self._original_keyboard_layout_is_desktop = (
            command_manager.get_manager().get_keyboard_layout_is_desktop()
        )
        # set_keyboard_layout_is_desktop() already applied the overrides and updated the grabs.
        self._populate_keybindings()
        if self._current_category and self._current_category in self._categories:
            self._populate_category_detail(self._current_category)
        self._modified_keybindings.clear()
        self._has_unsaved_changes = False
        self.refresh()

    def revert_changes(self) -> None:
        """Revert keyboard layout and orca modifier to their original values."""

        current_is_desktop = command_manager.get_manager().get_keyboard_layout_is_desktop()
        if current_is_desktop != self._original_keyboard_layout_is_desktop:
            command_manager.get_manager().load_keyboard_layout(
                self._original_keyboard_layout_is_desktop
            )

        orca_modifier_manager.get_manager().set_modifier_keys_override(None)

    def _populate_keybindings(self) -> None:
        """Build categories dictionary and populate the categories list."""

        if self._commands_categories_listbox is None:
            return

        self._categories.clear()
        for child in self._commands_categories_listbox.get_children():
            self._commands_categories_listbox.remove(child)

        app_name = AXObject.get_name(self._script.app) if self._script.app else ""
        if app_name:
            self._categories[app_name] = []

        all_commands = command_manager.get_manager().get_all_keyboard_commands()

        for cmd in all_commands:
            group_label = cmd.get_group_label()
            if group_label not in self._categories:
                self._categories[group_label] = []
            self._categories[group_label].append(cmd)

        if app_name and app_name in self._categories and not self._categories[app_name]:
            del self._categories[app_name]

        for commands in self._categories.values():
            commands.sort(key=lambda command: command.is_transient())

        orca_gui_helpers.set_listbox_separators(self._commands_categories_listbox)

        # Custom sort: For app-specific, app name first then screen-reader management.
        # Otherwise, screen-reader management first. After that, sort alphabetically.
        def sort_key(category_name):
            if app_name and category_name == app_name:
                return (0, category_name)
            if category_name == guilabels.KB_GROUP_SCREEN_READER_MANAGEMENT:
                if app_name and app_name in self._categories:
                    return (1, category_name)
                return (0, category_name)
            return (2, category_name)

        sorted_categories = sorted(self._categories.keys(), key=sort_key)
        for category_name in sorted_categories:
            self._add_stack_category_row(
                self._commands_categories_listbox,
                category_name,
                category=category_name,
            )

        self._commands_categories_listbox.show_all()

    def _on_category_activated(self, row: Gtk.ListBoxRow) -> None:
        """Handle category selection - navigate to detail page."""

        if not isinstance(row, preferences_grid_base.CategoryListBoxRow):
            return

        category_name = row.category
        if not category_name or category_name not in self._categories:
            return

        self._current_category = category_name
        self._populate_category_detail(category_name)
        self._show_stack_detail()

    def on_becoming_visible(self) -> None:
        """Reset to the categories view when this grid becomes visible."""

        self.reload()
        self._show_stack_categories()
        if self._commands_categories_listbox:
            self._commands_categories_listbox.grab_focus()

    def _show_stack_categories(self) -> None:
        """Switch to categories view and update title to main page."""

        super()._show_stack_categories()
        if self._title_change_callback:
            self._title_change_callback(guilabels.COMMANDS)

    def _show_stack_detail(self) -> None:
        """Switch to detail view and update title to category name."""

        super()._show_stack_detail()
        if self._current_category:
            if self._title_change_callback:
                self._title_change_callback(self._current_category)
            if self._commands_detail_listbox:
                self._commands_detail_listbox.get_accessible().set_name(self._current_category)

    # pylint: disable=no-member

    def _populate_category_detail(self, category_name: str) -> None:
        """Populate the detail page with keybindings for the given category."""

        if self._commands_detail_listbox is None or category_name not in self._categories:
            return

        for child in self._commands_detail_listbox.get_children():
            self._commands_detail_listbox.remove(child)

        commands = self._categories[category_name]
        for i, cmd in enumerate(commands):
            row = CommandListBoxRow()
            row.set_activatable(True)

            outer_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)

            if i > 0:
                separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
                outer_vbox.pack_start(separator, False, False, 0)

            vbox = orca_gui_helpers.create_vertical_row_box()

            description = cmd.get_description() or cmd.get_name()
            desc_label = Gtk.Label(label=description, xalign=0)
            desc_label.set_line_wrap(True)
            desc_label.set_hexpand(True)
            vbox.pack_start(desc_label, False, False, 0)

            binding = cmd.get_keybinding()
            binding_text = self._format_keybinding_text(binding) or ""

            binding_label = Gtk.Label(xalign=0)
            binding_label.set_markup(f"<small>{GLib.markup_escape_text(binding_text, -1)}</small>")
            binding_label.get_style_context().add_class("dim-label")
            vbox.pack_start(binding_label, False, False, 0)

            outer_vbox.pack_start(vbox, False, False, 0)
            row.add(outer_vbox)

            row.command = cmd
            row.vbox = vbox
            row.binding_label = binding_label
            row.show_all()
            self._commands_detail_listbox.add(row)

        self._commands_detail_listbox.show_all()

    def _on_keybinding_activated(self, _listbox: Gtk.ListBox, row: Gtk.ListBoxRow) -> None:
        """Handle keybinding row activation - start inline editing."""

        if self._keybinding_being_edited is not None:
            return

        if not isinstance(row, CommandListBoxRow):
            return

        command = row.command
        if not command:
            return

        vbox = row.vbox
        binding_label = row.binding_label
        if not vbox or not binding_label:
            return

        self._start_inline_editing(row, command, vbox, binding_label)

    def _create_capture_entry(self, command: KeyboardCommand) -> Gtk.Entry:
        """Creates and returns an entry widget configured for key capture."""

        entry = Gtk.Entry()
        entry.set_alignment(0.0)
        binding = command.get_keybinding()
        if binding and binding.keysymstring:
            current_text = self._format_keybinding_text(binding)
            entry.set_text(current_text or "")
        else:
            entry.set_text("")
        return entry

    def _start_inline_editing(
        self,
        row: Gtk.ListBoxRow,
        command: KeyboardCommand,
        vbox: Gtk.Box,
        binding_label: Gtk.Label,
    ) -> None:
        """Start inline editing of a keybinding."""

        binding_label.hide()

        capture_entry = self._create_capture_entry(command)
        vbox.pack_start(capture_entry, False, False, 0)
        capture_entry.show()

        def swap_widgets():
            vbox.remove(binding_label)
            capture_entry.grab_focus()
            return False

        GLib.idle_add(swap_widgets)

        self._captured_key = ("", 0, 0)
        self._orca_modifier_pressed_during_capture = False
        self._keybinding_being_edited = command.get_name()

        def on_key_release(_widget: Gtk.Widget, event: Gdk.EventKey) -> bool:
            event_string = Gdk.keyval_name(event.keyval)
            orca_mods = orca_modifier_manager.get_manager().get_orca_modifier_keys()
            if event_string in orca_mods:
                self._orca_modifier_pressed_during_capture = False
            return False

        def on_key_press(_widget: Gtk.Widget, event: Gdk.EventKey) -> bool:
            if event.keyval == Gdk.KEY_Escape and not self._orca_modifier_pressed_during_capture:
                self._finish_inline_editing(
                    row,
                    command,
                    vbox,
                    capture_entry,
                    canceled=True,
                )
                return True

            if event.keyval == Gdk.KEY_Return and not self._orca_modifier_pressed_during_capture:
                if self._captured_key[0]:
                    key_name, modifiers, click_count = self._captured_key
                    handler_name = command.get_name()
                    description_dup = self._find_duplicate_binding(
                        key_name,
                        modifiers,
                        click_count,
                        handler_name,
                        command.get_group_label(),
                    )
                    if description_dup:

                        def present_duplicate_error():
                            presentation_manager.get_manager().present_message(
                                messages.KB_ALREADY_BOUND % description_dup,
                            )
                            return False

                        GLib.idle_add(present_duplicate_error)
                        capture_entry.set_text("")
                        self._captured_key = ("", 0, 0)
                        return True

                self._finish_inline_editing(
                    row,
                    command,
                    vbox,
                    capture_entry,
                    canceled=False,
                )
                return True

            key_processed = self._process_key_captured(event)
            if not key_processed or not self._captured_key[0]:
                return True

            key_name, modifiers, click_count = self._captured_key

            if key_name in ["Delete", "BackSpace"] and not modifiers:
                capture_entry.set_text("")
                self._captured_key = ("", 0, 0)

                def present_delete_message():
                    presentation_manager.get_manager().present_message(messages.KB_DELETED)
                    return False

                GLib.idle_add(present_delete_message)
                return True

            modifier_names = keybindings.get_modifier_names(modifiers)
            click_count_string = keynames.get_click_count_string(click_count)
            if click_count_string:
                click_count_string = f" ({click_count_string})"
            new_string = modifier_names + key_name + click_count_string
            capture_entry.set_text(new_string)

            def present_message_after_keypress():
                presentation_manager.get_manager().present_message(
                    messages.KB_CAPTURED % new_string,
                )
                return False

            GLib.idle_add(present_message_after_keypress)

            return True

        capture_entry.connect("key-press-event", on_key_press)
        capture_entry.connect("key-release-event", on_key_release)
        row.capture_entry = capture_entry

        script = script_manager.get_manager().get_active_script()
        assert script
        presentation_manager.get_manager().present_message(messages.KB_ENTER_NEW_KEY)
        self._saved_commands = command_manager.get_manager().get_keyboard_commands()
        orca_modifier_manager.get_manager().remove_grabs_for_orca_modifiers()
        command_manager.get_manager().set_active_commands({}, "Capturing keys")
        ax_device_manager.get_manager().unmap_all_modifiers()

    def _finish_inline_editing(
        self,
        row: Gtk.ListBoxRow,
        command: KeyboardCommand,
        vbox: Gtk.Box,
        capture_entry: Gtk.Entry,
        canceled: bool,
    ) -> None:
        """Finish inline editing of a keybinding."""

        # Update keybinding before restoring commands so grabs match actual keybindings.
        # Otherwise a deleted binding's grab remains, consuming keystrokes without executing
        # anything (is_active() returns False when keybinding is None).
        if not canceled:
            captured_text = capture_entry.get_text().strip()
            script_manager.get_manager().get_active_script()
            handler_name = command.get_name()
            if not captured_text:
                command.set_keybinding(None)
                self._modified_keybindings[handler_name] = None
                self._has_unsaved_changes = True

                def present_delete_confirmation():
                    presentation_manager.get_manager().present_message(
                        messages.KB_DELETED_CONFIRMATION,
                    )
                    return False

                GLib.idle_add(present_delete_confirmation)
            else:
                key_name, modifiers, click_count = self._captured_key
                if key_name:
                    new_kb = keybindings.KeyBinding(key_name, modifiers, click_count)
                    command.set_keybinding(new_kb)
                    self._modified_keybindings[handler_name] = new_kb
                    self._has_unsaved_changes = True

                    def present_confirmation():
                        msg = messages.KB_CAPTURED_CONFIRMATION % captured_text
                        presentation_manager.get_manager().present_message(msg)
                        return False

                    GLib.idle_add(present_confirmation)

        command_manager.get_manager().set_active_commands(
            self._saved_commands,
            "Done capturing keys",
        )
        orca_modifier_manager.get_manager().add_grabs_for_orca_modifiers()

        binding = command.get_keybinding()
        binding_text = self._format_keybinding_text(binding) or ""
        new_label = Gtk.Label(xalign=0)
        new_label.set_markup(
            f"<small>{GLib.markup_escape_text(binding_text, -1)}</small>",
        )
        new_label.get_style_context().add_class("dim-label")
        row.binding_label = new_label

        capture_entry.hide()
        vbox.pack_start(new_label, False, False, 0)
        new_label.show()

        def swap_widgets():
            vbox.remove(capture_entry)
            row.grab_focus()
            return False

        GLib.idle_add(swap_widgets)

        self._captured_key = ("", 0, 0)
        self._keybinding_being_edited = None

    def _find_duplicate_binding(
        self,
        key_name: str,
        modifiers: int,
        click_count: int,
        exclude_handler: str | None = None,
        group_label: str = "",
    ) -> str | None:
        """Find if a keybinding is already used and return its description."""

        mgr = command_manager.get_manager()
        for category_label, category_commands in self._categories.items():
            if group_label and mgr.are_groups_exclusive(group_label, category_label):
                continue

            for cmd in category_commands:
                if exclude_handler and cmd.get_name() == exclude_handler:
                    continue

                binding = cmd.get_keybinding()
                if not binding or not binding.keysymstring:
                    continue

                if (
                    binding.keysymstring == key_name
                    and binding.modifiers == modifiers
                    and binding.click_count == click_count
                ):
                    return cmd.get_description() or cmd.get_name()

        return None

    def _process_key_captured(self, event: Gdk.EventKey) -> bool:
        """Process a captured key press event."""

        keycode = event.hardware_keycode
        keymap = Gdk.Keymap.get_default()  # pylint: disable=no-value-for-parameter
        entries_for_keycode = keymap.get_entries_for_keycode(keycode)
        entries = entries_for_keycode[-1]
        event_string = Gdk.keyval_name(entries[0])
        event_state = event.state

        orca_mods = orca_modifier_manager.get_manager().get_orca_modifier_keys()
        if event_string in orca_mods:
            self._orca_modifier_pressed_during_capture = True
            self._captured_key = ("", keybindings.ORCA_MODIFIER_MASK, 0)
            return False

        modifier_keys = [
            "Alt_L",
            "Alt_R",
            "Control_L",
            "Control_R",
            "Shift_L",
            "Shift_R",
            "Meta_L",
            "Meta_R",
            "Super_L",
            "Super_R",
            "Num_Lock",
            "Caps_Lock",
            "Shift_Lock",
            "ISO_Level3_Shift",
        ]
        if event_string in modifier_keys:
            return False

        event_state = event_state & keybindings.NON_LOCKING_MODIFIER_MASK

        # Return and Escape are used to confirm/cancel editing, not as captured keys
        # Return False to let GTK process them normally
        if event_string in ["Return", "Escape"] and not self._orca_modifier_pressed_during_capture:
            return False

        if not self._captured_key[0]:
            # Preserve Orca modifier if it was already captured.
            if self._captured_key[1] & keybindings.ORCA_MODIFIER_MASK:
                event_state |= keybindings.ORCA_MODIFIER_MASK
            self._captured_key = (event_string, event_state, 1)
            return True

        string, modifiers, click_count = self._captured_key

        # Preserve Orca modifier from previous key if present before comparing
        if modifiers & keybindings.ORCA_MODIFIER_MASK:
            event_state |= keybindings.ORCA_MODIFIER_MASK

        if string != event_string or modifiers != event_state:
            self._captured_key = (event_string, event_state, 1)
            return True

        # Same key pressed again - increment click count
        self._captured_key = (event_string, event_state, click_count + 1)
        return True

    # pylint: enable=no-member

    def save_settings(  # pylint: disable=too-many-locals, too-many-branches
        self,
        profile: str = "",
        app_name: str = "",
    ) -> tuple[dict[str, int | list[str]], dict[str, list[list[Any]]]]:
        """Save settings and return (general_settings, keybindings) tuple."""

        general: dict[str, int | list[str]] = {}
        bindings: dict[str, list[list[Any]]] = {}

        layout_value = command_manager.get_manager().get_keyboard_layout_value()
        if self.keyboard_layout_combo is not None:
            layout_iter = self.keyboard_layout_combo.get_active_iter()
            if layout_iter is not None:
                layout_value = self.keyboard_layout_combo.get_model().get_value(layout_iter, 1)
        general["keyboard-layout"] = layout_value

        is_desktop = layout_value == command_manager.KeyboardLayout.DESKTOP.value
        if self._modifier_switches:
            modifier_keys = self._get_selected_modifier_keys()
            if modifier_keys:
                if is_desktop:
                    general["desktop-modifier-keys"] = modifier_keys
                else:
                    general["laptop-modifier-keys"] = modifier_keys

        parent_overrides: dict[str, list[list[str]]] = {}
        if profile and (profile != "default" or app_name):
            registry = gsettings_registry.get_registry()
            if profile != "default":
                parent_overrides |= registry.get_keybindings("default", "")
            if app_name:
                parent_overrides |= registry.get_keybindings(profile, "")

        for category_commands in self._categories.values():
            for cmd in category_commands:
                handler_name = cmd.get_name()
                if command_manager.get_manager().get_keyboard_command(handler_name) is None:
                    continue
                if handler_name in self._modified_keybindings:
                    current_kb = self._modified_keybindings[handler_name]
                else:
                    current_kb = cmd.get_keybinding()

                current_text = self._format_keybinding_text(current_kb)

                if handler_name in parent_overrides:
                    parent_text = self._format_binding_data_text(parent_overrides[handler_name])
                else:
                    parent_text = self._format_keybinding_text(
                        cmd.get_default_keybinding(
                            command_manager.get_manager().is_desktop_layout(),
                        ),
                    )

                if current_text != parent_text:
                    msg = (
                        f"KEYBINDINGS GRID: Saving {handler_name}: '{current_text}' "
                        f"(parent '{parent_text}')"
                    )
                    debug.print_message(debug.LEVEL_INFO, msg, True)
                    if current_kb and current_kb.keysymstring:
                        binding_data = [
                            current_kb.keysymstring,
                            str(current_kb.modifier_mask),
                            str(current_kb.modifiers),
                            str(current_kb.click_count),
                        ]
                        bindings[handler_name] = [binding_data]
                    elif parent_text is not None:
                        bindings[handler_name] = []

        self._modified_keybindings.clear()
        self._has_unsaved_changes = False

        if profile:
            registry = gsettings_registry.get_registry()
            skip = not app_name and profile == "default"
            registry.save_schema("keybindings", general, profile, app_name, skip)
            kb_gs = registry.get_settings("keybindings", profile, "keybindings", app_name)
            if kb_gs is not None:
                if bindings:
                    kb_gs.set_value("entries", GLib.Variant("a{saas}", bindings))
                elif self._categories and kb_gs.get_user_value("entries") is not None:
                    kb_gs.reset("entries")

        return general, bindings

    def refresh(self) -> None:
        """Refresh the keyboard layout and orca modifier displays."""

        self._initializing = True

        if self.keyboard_layout_combo is not None:
            current_layout = command_manager.get_manager().get_keyboard_layout_value()
            model = self.keyboard_layout_combo.get_model()
            if model:
                for i, row in enumerate(model):
                    if row[1] == current_layout:
                        self.keyboard_layout_combo.set_active(i)
                        break

        if self._modifier_switches:
            layout_value = command_manager.get_manager().get_keyboard_layout_value()
            is_desktop = layout_value == command_manager.KeyboardLayout.DESKTOP.value
            app_name = AXObject.get_name(self._script.app) if self._script.app else ""
            modifier_keys = command_manager.get_manager().get_modifier_keys_for_layout(
                is_desktop,
                app_name,
            )
            for primary_key, switch in self._modifier_switches.items():
                switch.set_active(primary_key in modifier_keys)

        self._initializing = False

    def _on_keyboard_layout_changed(self, combo: Gtk.ComboBox) -> None:
        """Handle keyboard layout changes."""

        if self._initializing:
            return

        tree_iter = combo.get_active_iter()
        if tree_iter is not None:
            model = combo.get_model()
            layout_value = model.get_value(tree_iter, 1)

            is_desktop = layout_value == command_manager.KeyboardLayout.DESKTOP.value
            command_manager.get_manager().load_keyboard_layout(is_desktop)

            if self._modifier_switches:
                self._initializing = True
                app_name = AXObject.get_name(self._script.app) if self._script.app else ""
                saved_keys = command_manager.get_manager().get_modifier_keys_for_layout(
                    is_desktop,
                    app_name,
                )
                for primary_key, switch in self._modifier_switches.items():
                    switch.set_active(primary_key in saved_keys)
                self._initializing = False
                orca_modifier_manager.get_manager().set_modifier_keys_override(
                    self._get_selected_modifier_keys(),
                )

        # load_keyboard_layout() already applied the overrides and updated the grabs.
        self._populate_keybindings()
        self._has_unsaved_changes = True

    def _get_selected_modifier_keys(self) -> list[str]:
        """Returns the list of modifier keys from the current switch states."""

        keys: list[str] = []
        for _label, key_names in self._MODIFIER_KEY_OPTIONS:
            switch = self._modifier_switches.get(key_names[0])
            if switch is not None and switch.get_active():
                keys.extend(key_names)
        return keys

    def _on_modifier_switch_toggled(self, switch: Gtk.Switch, _pspec: Any) -> None:
        """Handle modifier key switch toggles."""

        if self._initializing:
            return

        selected = self._get_selected_modifier_keys()
        if not selected:
            switch.set_active(True)
            return

        orca_modifier_manager.get_manager().set_modifier_keys_override(selected)
        self._has_unsaved_changes = True

    def _format_keybinding_text(self, kb: keybindings.KeyBinding | None) -> str | None:
        """Format a keybinding as text for display."""

        if not kb or not kb.keysymstring:
            return None

        click_count_str = keynames.get_click_count_string(kb.click_count)
        if click_count_str:
            click_count_str = f" ({click_count_str})"

        return keybindings.get_modifier_names(kb.modifiers) + kb.keysymstring + click_count_str

    @staticmethod
    def _format_binding_data_text(binding_data: list[list[str]]) -> str | None:
        """Format raw dconf binding data as text, matching _format_keybinding_text output."""

        if not binding_data:
            return None
        entry = binding_data[0]
        if len(entry) < 4 or not entry[0]:
            return None
        click_count_str = keynames.get_click_count_string(int(entry[3]))
        if click_count_str:
            click_count_str = f" ({click_count_str})"
        return keybindings.get_modifier_names(int(entry[2])) + entry[0] + click_count_str



# Orca
#
# Copyright 2023-2024 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Utilities for accessible applications."""

from __future__ import annotations

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi, GLib

from . import ax_cache_manager, debug
from .ax_object import AXObject


class _AXUtilitiesApplicationCache:
    """Provides application-specific access to manager-backed cached values."""

    APP_FOR_OBJECT = "AXUtilitiesApplication.app-for-object"
    _CACHE_CLEAR_INTERVAL_SECONDS = 120

    def __init__(self) -> None:
        self._manager = ax_cache_manager.get_manager()
        self._manager.register_cache(
            self,
            self.APP_FOR_OBJECT,
            lifetime=ax_cache_manager.Lifetime.PROCESS,
            clear_on_demand=ax_cache_manager.ClearPolicy.PRESERVE,
            clear_interval_seconds=self._CACHE_CLEAR_INTERVAL_SECONDS,
        )
        self._apps_for_objects = self._manager.get_cache(self, self.APP_FOR_OBJECT)

    def get_application(self, obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the cached application for obj."""

        if self._apps_for_objects is None:
            return None
        return self._apps_for_objects.get(ax_cache_manager.get_object_key(obj), None)

    def set_application(self, obj: Atspi.Accessible, app: Atspi.Accessible) -> None:
        """Stores the application for obj."""

        if self._apps_for_objects is not None:
            self._apps_for_objects.put(ax_cache_manager.get_object_key(obj), app)


class AXUtilitiesApplication:
    """Utilities for accessible applications."""

    _CACHE = _AXUtilitiesApplicationCache()

    _mutter_x11_frames: Atspi.Accessible | None = None
    _mutter_x11_frames_checked: bool = False

    @staticmethod
    def application_as_string(obj: Atspi.Accessible) -> str:
        """Returns the application details of obj as a string."""

        app = AXUtilitiesApplication.get_application(obj)
        if app is None:
            return ""

        string = (
            f"{AXObject.get_name(app)} "
            f"({AXUtilitiesApplication.get_application_toolkit_name(obj)} "
            f"{AXUtilitiesApplication.get_application_toolkit_version(obj)})"
        )
        return string

    @staticmethod
    def get_all_applications(
        must_have_window: bool = False,
        exclude_unresponsive: bool = False,
        is_debug: bool = False,
    ) -> list[Atspi.Accessible]:
        """Returns a list of running applications known to Atspi."""

        desktop = AXUtilitiesApplication.get_desktop()
        if desktop is None:
            return []

        def pred(obj: Atspi.Accessible) -> bool:
            if exclude_unresponsive and AXUtilitiesApplication.is_application_unresponsive(obj):
                return False
            if AXUtilitiesApplication.is_mutter_x11_frames(obj):
                return is_debug
            if must_have_window:
                return AXObject.get_child_count(obj) > 0
            return True

        result = list(AXObject.iter_children(desktop, pred))
        if not AXUtilitiesApplication._mutter_x11_frames_checked:
            AXUtilitiesApplication._mutter_x11_frames_checked = True
            if AXUtilitiesApplication._mutter_x11_frames is None:
                msg = "AXUtilitiesApplication: mutter-x11-frames not found on this desktop"
                debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    @staticmethod
    def get_application(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the accessible application associated with obj"""

        if not AXObject.is_valid(obj):
            return None

        cached = AXUtilitiesApplication._CACHE.get_application(obj)
        if cached is not None:
            return cached

        parent = None
        if AXObject.get_role(obj) != Atspi.Role.APPLICATION:
            parent = AXObject.get_parent(obj)
        if parent is not None:
            cached = AXUtilitiesApplication._CACHE.get_application(parent)
            if cached is not None:
                AXUtilitiesApplication._CACHE.set_application(obj, cached)
                return cached

        try:
            app = Atspi.Accessible.get_application(obj)
        except GLib.GError as error:
            msg = f"AXUtilitiesApplication: Exception in get_application: {error}"
            AXObject.handle_error(obj, error, msg)
            return None
        if app is not None:
            AXUtilitiesApplication._CACHE.set_application(obj, app)
            if parent is not None:
                AXUtilitiesApplication._CACHE.set_application(parent, app)
        return app

    @staticmethod
    def get_application_toolkit_name(obj: Atspi.Accessible) -> str:
        """Returns the toolkit name reported for obj's application."""

        app = AXUtilitiesApplication.get_application(obj)
        if app is None:
            return ""

        try:
            name = Atspi.Accessible.get_toolkit_name(app)
        except GLib.GError as error:
            msg = f"AXUtilitiesApplication: Exception in get_application_toolkit_name: {error}"
            AXObject.handle_error(app, error, msg)
            return ""

        return name or ""

    @staticmethod
    def get_application_toolkit_version(obj: Atspi.Accessible) -> str:
        """Returns the toolkit version reported for obj's application."""

        app = AXUtilitiesApplication.get_application(obj)
        if app is None:
            return ""

        try:
            version = Atspi.Accessible.get_toolkit_version(app)
        except GLib.GError as error:
            msg = f"AXUtilitiesApplication: Exception in get_application_toolkit_version: {error}"
            AXObject.handle_error(app, error, msg)
            return ""

        return version or ""

    @staticmethod
    def get_application_with_pid(pid: int) -> Atspi.Accessible | None:
        """Returns the accessible application with the specified pid"""

        applications = AXUtilitiesApplication.get_all_applications()
        for child in applications:
            if AXUtilitiesApplication.get_process_id(child) == pid:
                return child

        tokens = ["WARNING: app with pid", pid, "is not in the accessible desktop"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return None

    @staticmethod
    def get_desktop() -> Atspi.Accessible | None:
        """Returns the accessible desktop"""

        try:
            desktop = Atspi.get_desktop(0)
        except GLib.GError as error:
            tokens = ["ERROR: Exception getting desktop from Atspi:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        return desktop

    @staticmethod
    def get_process_id(obj: Atspi.Accessible) -> int:
        """Returns the process id associated with obj"""

        try:
            pid = Atspi.Accessible.get_process_id(obj)
        except GLib.GError as error:
            msg = f"AXUtilitiesApplication: Exception in get_process_id: {error}"
            AXObject.handle_error(obj, error, msg)
            return -1

        return pid

    @staticmethod
    def is_mutter_x11_frames(app: Atspi.Accessible) -> bool:
        """Returns True if app is the mutter-x11-frames application."""

        if app is None:
            return False
        if AXUtilitiesApplication._mutter_x11_frames is not None:
            return app == AXUtilitiesApplication._mutter_x11_frames
        if AXUtilitiesApplication._mutter_x11_frames_checked:
            return False
        if AXObject.get_name(app) == "mutter-x11-frames":
            tokens = ["AXUtilitiesApplication: Caching mutter-x11-frames app:", app]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            AXUtilitiesApplication._mutter_x11_frames = app
            return True
        return False

    @staticmethod
    def is_application_in_desktop(app: Atspi.Accessible) -> bool:
        """Returns true if app is known to Atspi"""

        desktop = AXUtilitiesApplication.get_desktop()
        parent = AXObject.get_parent(app)
        if desktop is not None and parent == desktop:
            return True

        tokens = ["WARNING:", app, "with parent", parent, "is not in the accessible desktop."]
        if debug.debugLevel <= debug.LEVEL_INFO:
            tokens.append(AXUtilitiesApplication.application_as_string(app))
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        # Qt 5/6 and GTK 4 apps can fail to report their parent even though the desktop lists them
        # among its children, so for them fall back to the slower scan of the desktop's children.
        if parent is None:
            toolkit = AXUtilitiesApplication.get_application_toolkit_name(app).lower()
            major = AXUtilitiesApplication.get_application_toolkit_version(app).split(".")[0]
            if (toolkit, major) in (
                ("qt", "5"),
                ("qt", "6"),
                ("gtk", "4"),
            ) and app in AXUtilitiesApplication.get_all_applications():
                return True

        return False

    @staticmethod
    def is_application_unresponsive(app: Atspi.Accessible) -> bool:
        """Returns true if app's process is known to be unresponsive."""

        pid = AXUtilitiesApplication.get_process_id(app)
        try:
            with open(f"/proc/{pid}/status", encoding="utf-8") as f:
                state = ""
                for line in f:
                    if line.startswith("State:"):
                        state = line.split()[1]
                        break
        except (OSError, IndexError) as error:
            tokens = [f"AXUtilitiesApplication: Exception checking state of pid {pid}: {error}"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if state == "Z":
            tokens = [f"AXUtilitiesApplication: pid {pid} is zombie process"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        if state == "T":
            tokens = [f"AXUtilitiesApplication: pid {pid} is suspended/stopped process"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return False



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2010 Orca Team.
# Copyright 2014-2015 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Custom script for Gecko."""

from __future__ import annotations

from typing import TYPE_CHECKING

from orca import debug, focus_manager
from orca.ax_utilities import AXUtilities
from orca.scripts import web

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi


class Script(web.ToolkitBridge):
    """Custom script for Gecko."""

    def _on_focused_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:focused accessibility events."""

        if AXUtilities.is_panel(event.source):
            if focus_manager.get_manager().focus_is_active_window():
                msg = "GECKO: Ignoring event believed to be noise."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

        if AXUtilities.is_frame(event.source):
            msg = "GECKO: Ignoring event believed to be noise."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        return super()._on_focused_changed(event)

    def _on_showing_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:showing accessibility events."""

        # TODO - JD: Is this workaround still needed? It is here because we normally get a
        # window:activate event when a context menu is shown, but not in the case of Gecko.
        if (
            event.detail1
            and AXUtilities.is_menu(event.source)
            and not self.utilities.in_document_content(event.source)
        ):
            msg = "GECKO: Setting locus of focus to newly shown menu."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            focus_manager.get_manager().set_locus_of_focus(event, event.source)
            return True

        return super()._on_showing_changed(event)
